@@ -239,7 +239,8 @@ def path_to_tree_path(repopath, path, tree_encoding=DEFAULT_ENCODING):
         if sys.platform == "win32":
             path = os.path.abspath(path)
 
-        path = Path(path).resolve()
+        path = Path(path)
+        resolved_path = path.resolve()
 
         # Resolve and abspath seems to behave differently regarding symlinks,
         # as we are doing abspath on the file path, we need to do the same on
@@ -249,7 +250,16 @@ def path_to_tree_path(repopath, path, tree_encoding=DEFAULT_ENCODING):
 
         repopath = Path(repopath).resolve()
 
-        relpath = path.relative_to(repopath)
+        try:
+            relpath = resolved_path.relative_to(repopath)
+        except ValueError:
+            # If path is a symlink that points to a file outside the repo, we
+            # want the relpath for the link itself, not the resolved target
+            if path.is_symlink():
+                parent = path.parent.resolve()
+                relpath = (parent / path.name).relative_to(repopath)
+            else:
+                raise
         if sys.platform == "win32":
             return str(relpath).replace(os.path.sep, "/").encode(tree_encoding)
         else:
@@ -465,8 +475,8 @@ def clone(
             depth=depth,
             **kwargs
         )
-        # TODO(jelmer): Support symref capability,
-        # https://github.com/jelmer/dulwich/issues/485
+        for key, target in fetch_result.symrefs.items():
+            r.refs.set_symbolic_ref(key, target)
         try:
             head = r[fetch_result.refs[b"HEAD"]]
         except KeyError:
@@ -1055,6 +1065,8 @@ def push(
 
     # Open the repo
     with open_repo_closing(repo) as r:
+        if refspecs is None:
+            refspecs = [active_branch(r)]
         (remote_name, remote_location) = get_remote_repo(r, remote_location)
 
         # Get the client and path
